import logging
import requests

from gracedb_sdk import Client

GRACEDB_BASE_URL = 'https://gracedb.ligo.org'
GRACEDB_EVENT_URL = f'{GRACEDB_BASE_URL}/superevents'
GRACEDB_API_URL = f'{GRACEDB_BASE_URL}/api/superevents/files'
SKIP_BASE_URL = 'http://skip.dev.hop.scimma.org'
SKIP_API_URL = f'{SKIP_BASE_URL}/api/events'

logger = logging.getLogger(__name__)


class GravitationalWaveClient:

    def __init__(self):
        self.gracedb_client = Client()

    def _get_fwhm_url(self, superevent_id):
        fwhm_data = {}

        try:
            fwhm_url = self.gracedb_client.superevents[superevent_id].files.get()['bayestar.volume.png']
            fwhm_data['fwhm_url'] = fwhm_url
        except Exception as e:
            logger.warn(f'Unable to retrieve FWHM URL for {superevent_id}: {e}')

        return fwhm_data

    def _get_skymap_url(self, superevent_id):
        skymap_data = {}

        try:
            skymap_url = self.gracedb_client.superevents[superevent_id].files.get()['bayestar.png']
            skymap_data['skymap_url'] = skymap_url
        except Exception as e:
            logger.warn(f'Unable to retrieve Skymap URL for {superevent_id}: {e}')

        return skymap_data

    def get_superevent_data(self, superevent_id: str) -> dict:
        url = requests.get(f'{SKIP_API_URL}/?identifier={superevent_id}').json()['results'][0]['event_detail']
        response = requests.get(url).json()
        superevent_data = {}
        superevent_data['event_data'] = response['event_attributes'][0]
        superevent_data['alerts'] = response['alerts']
        superevent_data['url'] = f'{GRACEDB_EVENT_URL}/{superevent_id}/'
        return superevent_data

    def get_additional_context_data(self, superevent_id: str) -> dict:
        additional_context_data = {}
        additional_context_data.update(self._get_fwhm_url(superevent_id))
        additional_context_data.update(self._get_skymap_url(superevent_id))
        return additional_context_data



from urllib3.response import HTTPResponse

from gracedb_sdk import Client
import voeventparse as vp


GRACEDB_BASE_URL = 'https://gracedb.ligo.org'
GRACEDB_EVENT_URL = f'{GRACEDB_BASE_URL}/superevents'
GRACEDB_API_URL = f'{GRACEDB_BASE_URL}/api/superevents'


@DeprecationWarning
class GraceDBClient:

    def __init__(self, *args, **kwargs):
        self.gracedb_client = Client()

    def get_superevent_data(self, superevent_id: str) -> dict:
        print('this isn\'t great')
        try:
            gracedb_superevent = self.gracedb_client.superevents[superevent_id]
            gracedb_data = gracedb_superevent.get()
            latest_voevent = gracedb_superevent.voevents.get()[-1]
            voevent_file = gracedb_superevent.files[latest_voevent['filename']].get()

            superevent_data = {}
            superevent_data['update_version'] = latest_voevent['filename'].split('-')[1]
            superevent_data['far'] = gracedb_data['far']
            superevent_data['instruments'] = gracedb_data['preferred_event_data']['instruments']
            superevent_data.update(self.get_data_from_voevent(voevent_file))
            return superevent_data
        except Exception as e:
            print(f'exception {e}')
            return {}

    # TODO: error handle the crap out of this
    def get_data_from_voevent(self, voevent_file: HTTPResponse) -> dict:
        voevent = vp.load(voevent_file)
        voevent_data = {}

        classification_group = voevent.What.find(".//Group[@type='Classification']")
        for datum in ['BNS', 'NSBH', 'BBH', 'MassGap', 'Terrestrial']:
            datum_element = classification_group.find(f".//Param[@name='{datum}']")
            if datum_element is not None:
                voevent_data[datum] = datum_element.attrib['value']

        properties_group = voevent.What.find(".//Group[@type='Properties']")
        for datum in ['HasNS', 'HasRemnant']:
            datum_element = properties_group.find(f".//Param[@name='{datum}']")
            if datum_element is not None:
                voevent_data[datum] = datum_element.attrib['value']

        return voevent_data



from django.views.generic import DetailView, ListView

from rest_framework import permissions, viewsets

from tom_nonlocalizedevents.superevent_clients.gravitational_wave import GravitationalWaveClient

from .models import EventCandidate, EventLocalization, Superevent
from .serializers import EventCandidateSerializer, EventLocalizationSerializer, SupereventSerializer


class NonlocalizedEventListView(ListView):
    """
    Unadorned Django ListView subclass for Superevent model.
    (To be updated when Superevent model is renamed to NonlocalizedEvent).
    """
    model = Superevent
    template_name = 'tom_nonlocalizedevents/index.html'


class NonlocalizedEventDetailView(DetailView):
    """
    Django DetailView subclass for SuperEvent model.
    (To be updated when Superevent model is renamed to NonlocalizedEvent).

    Has mechanism to supply templates specific to the type of NonlocalizedEvent
    (GW, GRB, Nutrino).
    """
    model = Superevent
    template_name = 'tom_nonlocalizedevents/detail.html'

    # TODO: consider combining these dictionaries
    template_mapping = {
        Superevent.SupereventType.GRAVITATIONAL_WAVE:
            'tom_nonlocalizedevents/superevent_detail/gravitational_wave.html',
        Superevent.SupereventType.GAMMA_RAY_BURST:
            'tom_nonlocalizedevents/superevent_detail/gamma_ray_burst.html',
        Superevent.SupereventType.NEUTRINO:
            'tom_nonlocalizedevents/superevent_detail/neutrino.html',
    }

    # A client in this context is the interface to the service providing event info.
    # (i.e GraceDB for gravitational wave events)
    client_mapping = {
        Superevent.SupereventType.GRAVITATIONAL_WAVE: GravitationalWaveClient(),
        Superevent.SupereventType.GAMMA_RAY_BURST: None,
        Superevent.SupereventType.NEUTRINO: None,
        Superevent.SupereventType.UNKNOWN: None,
    }

    def get_template_names(self):
        obj = self.get_object()
        return [self.template_mapping[obj.superevent_type]]

    # TODO: error handling
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        obj = self.get_object()
        superevent_client = self.client_mapping[obj.superevent_type]
        # TODO: should define superevent_client API (via ABC) for clients to implement
        if superevent_client is not None:
            context['superevent_data'] = superevent_client.get_superevent_data(obj.superevent_id)
            context.update(superevent_client.get_additional_context_data(obj.superevent_id))
        return context


# Django Rest Framework Views


class NonlocalizedEventViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint that allows Superevents to be viewed or edited.
    """
    queryset = Superevent.objects.all()
    serializer_class = SupereventSerializer
    permission_classes = []


class EventCandidateViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint for EventCandidate model.

    Implementation has changes for bulk_create and update/PATCH EventCandidate instances.
    """
    queryset = EventCandidate.objects.all()
    serializer_class = EventCandidateSerializer
    permission_classes = []  # TODO: re-implement auth permissions

    def get_serializer(self, *args, **kwargs):
        # In order to ensure the list_serializer_class is used for bulk_create, we check that the POST data is a list
        # and add `many = True` to the kwargs
        if isinstance(kwargs.get('data', {}), list):
            kwargs['many'] = True

        return super().get_serializer(*args, **kwargs)

    def update(self, request, *args, **kwargs):
        """Provide support for the PATCH HTTP verb to update individual model fields.

        An example request might look like:

            PATCH http://localhost:8000/api/eventcandidates/18/

        with a Request Body of:

            {
                "viability": false
            }

        """
        kwargs['partial'] = True
        return super().update(request, *args, **kwargs)


class EventLocalizationViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint that allows EventLocalizations to be viewed or edited.
    """
    queryset = EventLocalization.objects.all()
    serializer_class = EventLocalizationSerializer
    permission_classes = [permissions.IsAuthenticated]



from dateutil.parser import parse

from django import template

register = template.Library()


@register.filter
def percentage_filter(value):
    """
    Converts a decimal to a percentage with a single decimal place.
    """
    try:
        return f"{'%.1f' % (float(value) * 100)}%"
    except Exception:
        print(f'exception: {value}')
        return value


@register.inclusion_tag('tom_nonlocalizedevents/partials/alert_table.html', takes_context=True)
def alert_table(context):
    """
    Displays the alerts of an event.
    """
    alerts = [{'url': f'http://skip.dev.hop.scimma.org/api/v2/alerts/{alert["id"]}/',
               'identifier': f'GCN {alert["identifier"]}',
               'timestamp': str(parse(alert['timestamp']).date()),
               'from': alert['parsed_message']['from'].split(' at ', 1)[0],
               'subject': alert['parsed_message']['subject'].split(':', 1)[-1]
               } for alert in context['superevent_data']['alerts'][:8]
              if alert['parsed_message'].get('title', None) != 'GCN/LVC NOTICE']
    superevent_alert = None
    for alert in context['superevent_data']['alerts']:
        if alert['parsed_message'].get('title', None) == 'GCN/LVC NOTICE':
            superevent_alert = alert
            superevent_alert['timestamp'] = str(parse(superevent_alert['timestamp']).date())
            superevent_alert['url'] = context['superevent_data']['url']
            break

    return {
        'alerts': alerts,
        'superevent_alert': superevent_alert
    }
